﻿using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.ProgramSynthesis {
  /// <summary>
  /// Takes the rest of the top item on the type stack.
  /// </summary>
  /// <typeparam name="T"></typeparam>
  [StorableClass]
  public abstract class VectorButLastExpression<T> : StatelessExpression {
    protected VectorButLastExpression() { }
    [StorableConstructor]
    protected VectorButLastExpression(bool deserializing) : base(deserializing) { }

    protected bool IsNoop(IPushStack<IReadOnlyList<T>> vectorStack) {
      return vectorStack.IsEmpty || vectorStack.Top.Count < 2;
    }

    protected void Eval(IPushStack<IReadOnlyList<T>> vectorStack) {
      vectorStack.Top = vectorStack.Top
        .Take(vectorStack.Top.Count - 1)
        .ToList();
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.IntegerVector,
    "INTEGER[].BUTLAST",
    "Takes the rest of the top item on the INTEGER[] stack.")]
  public class IntegerVectorButLastExpression : VectorButLastExpression<long> {
    public IntegerVectorButLastExpression() { }
    [StorableConstructor]
    protected IntegerVectorButLastExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.IntegerVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.IntegerVectorStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.FloatVector,
    "FLOAT[].BUTLAST",
    "Takes the rest of the top item on the FLOAT[] stack.")]
  public class FloatVectorButLastExpression : VectorButLastExpression<double> {
    public FloatVectorButLastExpression() { }
    [StorableConstructor]
    protected FloatVectorButLastExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.FloatVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.FloatVectorStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.BooleanVector,
    "BOOLEAN[].BUTLAST",
    "Takes the rest of the top item on the BOOLEAN[] stack.")]
  public class BooleanVectorButLastExpression : VectorButLastExpression<bool> {
    public BooleanVectorButLastExpression() { }
    [StorableConstructor]
    protected BooleanVectorButLastExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.BooleanVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.BooleanVectorStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.StringVector,
    "STRING[].BUTLAST",
    "Takes the rest of the top item on the STRING[] stack.")]
  public class StringVectorButLastExpression : VectorButLastExpression<string> {
    public StringVectorButLastExpression() { }
    [StorableConstructor]
    protected StringVectorButLastExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.StringVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.StringVectorStack);
    }
  }
}
